/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.client;

import java.util.Arrays;
import javax.annotation.Nonnull;

import com.google.common.base.Strings;
import com.google.common.collect.ImmutableMap;
import net.minecraft.block.*;
import net.minecraft.block.material.Material;
import net.minecraft.client.renderer.ItemMeshDefinition;
import net.minecraft.client.renderer.block.model.ModelBakery;
import net.minecraft.client.renderer.block.model.ModelResourceLocation;
import net.minecraft.client.renderer.block.statemap.StateMap;
import net.minecraft.client.renderer.color.BlockColors;
import net.minecraft.client.renderer.color.ItemColors;
import net.minecraft.item.EnumDyeColor;
import net.minecraft.item.Item;
import net.minecraft.item.ItemBlock;
import net.minecraft.item.ItemStack;
import net.minecraft.util.ResourceLocation;
import net.minecraft.world.ColorizerFoliage;
import net.minecraft.world.ColorizerGrass;
import net.minecraft.world.biome.BiomeColorHelper;
import net.minecraftforge.client.event.ColorHandlerEvent;
import net.minecraftforge.client.event.ModelRegistryEvent;
import net.minecraftforge.client.model.ModelLoader;
import net.minecraftforge.fluids.BlockFluidBase;
import net.minecraftforge.fluids.capability.CapabilityFluidHandler;
import net.minecraftforge.fluids.capability.IFluidHandler;
import net.minecraftforge.fml.client.registry.ClientRegistry;
import net.minecraftforge.fml.common.Mod;
import net.minecraftforge.fml.common.eventhandler.SubscribeEvent;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import net.dries007.tfc.api.capability.IMoldHandler;
import net.dries007.tfc.api.registries.TFCRegistries;
import net.dries007.tfc.api.types.Metal;
import net.dries007.tfc.api.types.Ore;
import net.dries007.tfc.api.types.Rock;
import net.dries007.tfc.client.render.TESRChestTFC;
import net.dries007.tfc.client.render.TESRIngotPile;
import net.dries007.tfc.client.render.TESRPitKiln;
import net.dries007.tfc.client.render.TESRWorldItem;
import net.dries007.tfc.objects.Gem;
import net.dries007.tfc.objects.blocks.BlockSlabTFC;
import net.dries007.tfc.objects.blocks.BlocksTFC;
import net.dries007.tfc.objects.blocks.stone.BlockFarmlandTFC;
import net.dries007.tfc.objects.blocks.stone.BlockOreTFC;
import net.dries007.tfc.objects.blocks.stone.BlockRockVariant;
import net.dries007.tfc.objects.blocks.wood.BlockLeavesTFC;
import net.dries007.tfc.objects.blocks.wood.BlockLogTFC;
import net.dries007.tfc.objects.blocks.wood.BlockSaplingTFC;
import net.dries007.tfc.objects.items.ItemGem;
import net.dries007.tfc.objects.items.ItemGoldPan;
import net.dries007.tfc.objects.items.ItemsTFC;
import net.dries007.tfc.objects.items.ceramics.ItemMold;
import net.dries007.tfc.objects.items.metal.ItemOreTFC;
import net.dries007.tfc.objects.te.TEChestTFC;
import net.dries007.tfc.objects.te.TEIngotPile;
import net.dries007.tfc.objects.te.TEPitKiln;
import net.dries007.tfc.objects.te.TEWorldItem;

import static net.dries007.tfc.api.util.TFCConstants.MOD_ID;

@SideOnly(Side.CLIENT)
@Mod.EventBusSubscriber(value = Side.CLIENT, modid = MOD_ID)
public final class ClientRegisterEvents
{
    @SubscribeEvent
    @SuppressWarnings("ConstantConditions")
    public static void registerModels(ModelRegistryEvent event)
    {
        for (Item item : ItemsTFC.getAllSimpleItems())
            ModelLoader.setCustomModelResourceLocation(item, 0, new ModelResourceLocation(item.getRegistryName().toString()));

        for (EnumDyeColor color : EnumDyeColor.values())
        {
            ModelLoader.setCustomModelResourceLocation(ItemsTFC.CERAMICS_UNFIRED_VESSEL_GLAZED, color.func_176767_b(), new ModelResourceLocation(ItemsTFC.CERAMICS_UNFIRED_VESSEL_GLAZED.getRegistryName().toString()));
            ModelLoader.setCustomModelResourceLocation(ItemsTFC.CERAMICS_FIRED_VESSEL_GLAZED, color.func_176767_b(), new ModelResourceLocation(ItemsTFC.CERAMICS_FIRED_VESSEL_GLAZED.getRegistryName().toString()));
        }

        for (int meta = 0; meta < ItemGoldPan.TYPES.length; meta++)
            ModelLoader.setCustomModelResourceLocation(ItemsTFC.GOLDPAN, meta, new ModelResourceLocation(MOD_ID + ":goldpan/" + ItemGoldPan.TYPES[meta]));
        ModelLoader.registerItemVariants(ItemsTFC.GOLDPAN, Arrays.stream(ItemGoldPan.TYPES).map(e -> new ResourceLocation(MOD_ID, "goldpan/" + e)).toArray(ResourceLocation[]::new));

        ModelBakery.registerItemVariants(ItemMold.get(Metal.ItemType.INGOT), new ModelResourceLocation(ItemMold.get(Metal.ItemType.INGOT).getRegistryName() + "/unknown"));
        for (Metal.ItemType value : Metal.ItemType.values())
        {
            ItemMold item = ItemMold.get(value);
            if (item == null) continue;

            ModelLoader.setCustomMeshDefinition(item, new ItemMeshDefinition()
            {
                private ModelResourceLocation FALLBACK = new ModelResourceLocation(item.getRegistryName().toString() + "/empty");

                @Override
                @Nonnull
                public ModelResourceLocation func_178113_a(@Nonnull ItemStack stack)
                {
                    IFluidHandler cap = stack.getCapability(CapabilityFluidHandler.FLUID_HANDLER_CAPABILITY, null);
                    if (cap instanceof IMoldHandler)
                    {
                        Metal metal = ((IMoldHandler) cap).getMetal();
                        if (metal != null)
                        {
                            return new ModelResourceLocation(stack.func_77973_b().getRegistryName() + "/" + metal.getRegistryName().func_110623_a());
                        }
                    }
                    return FALLBACK;
                }
            });

            ModelBakery.registerItemVariants(item, new ModelResourceLocation(item.getRegistryName().toString() + "/empty"));
            ModelBakery.registerItemVariants(item, TFCRegistries.METALS.getValuesCollection()
                .stream()
                .filter(x -> item.type.hasMold && x.isToolMetal() && (x.getTier() == Metal.Tier.TIER_I || x.getTier() == Metal.Tier.TIER_II))
                .map(x -> new ModelResourceLocation(item.getRegistryName().toString() + "/" + x.getRegistryName().func_110623_a()))
                .toArray(ModelResourceLocation[]::new));
        }

        for (ItemGem item : ItemsTFC.getAllGemItems())
            for (Gem.Grade grade : Gem.Grade.values())
                registerEnumBasedMetaItems("gem", grade, item);

        for (ItemOreTFC item : ItemsTFC.getAllOreItems())
            if (item.ore.isGraded())
                for (Ore.Grade grade : Ore.Grade.values())
                    registerEnumBasedMetaItems("ore", grade, item);
            else
                registerEnumBasedMetaItems("ore", Ore.Grade.NORMAL, item);

        for (Block block : BlocksTFC.getAllFluidBlocks())
            ModelLoader.setCustomStateMapper(block, new StateMap.Builder().func_178442_a(BlockFluidBase.LEVEL).func_178441_a());

        for (Block block : BlocksTFC.getAllFenceGateBlocks())
            ModelLoader.setCustomStateMapper(block, new StateMap.Builder().func_178442_a(BlockFenceGate.field_176465_b).func_178441_a());

        for (Block block : BlocksTFC.getAllLeafBlocks())
            ModelLoader.setCustomStateMapper(block, new StateMap.Builder().func_178442_a(BlockLeaves.field_176237_a).func_178441_a());

        for (Block block : BlocksTFC.getAllOreBlocks())
            ModelLoader.setCustomStateMapper(block, new StateMap.Builder().func_178442_a(BlockOreTFC.GRADE).func_178441_a());

        for (Block block : BlocksTFC.getAllWallBlocks())
            ModelLoader.setCustomStateMapper(block, new StateMap.Builder().func_178442_a(BlockWall.field_176255_P).func_178441_a());

        for (Block block : BlocksTFC.getAllLogBlocks())
            ModelLoader.setCustomStateMapper(block, new StateMap.Builder().func_178442_a(BlockLogTFC.PLACED).func_178441_a());

        for (Block block : BlocksTFC.getAllSaplingBlocks())
            ModelLoader.setCustomStateMapper(block, new StateMap.Builder().func_178442_a(BlockSaplingTFC.STAGE).func_178441_a());

        for (Block block : BlocksTFC.getAllDoorBlocks())
            ModelLoader.setCustomStateMapper(block, new StateMap.Builder().func_178442_a(BlockDoor.field_176522_N).func_178441_a());

        for (Block block : BlocksTFC.getAllAnvils())
            ModelLoader.setCustomStateMapper(block, new StateMap.Builder().func_178441_a());

        for (Block block : BlocksTFC.getAllSheets())
            ModelLoader.setCustomStateMapper(block, new StateMap.Builder().func_178441_a());

        for (BlockSlabTFC.Half block : BlocksTFC.getAllSlabBlocks())
        {
            ModelLoader.setCustomStateMapper(block, new StateMap.Builder().func_178442_a(BlockSlabTFC.VARIANT).func_178441_a());
            ModelLoader.setCustomStateMapper(block.doubleSlab, new StateMap.Builder().func_178442_a(BlockSlabTFC.VARIANT).func_178441_a());
        }

        BlocksTFC.getAllBlockRockVariants().stream().filter(x -> x.type == Rock.Type.FARMLAND).forEach(e ->
            ModelLoader.setCustomStateMapper(e, new StateMap.Builder().func_178442_a(BlockFarmlandTFC.MOISTURE).func_178441_a())
        );

        for (Block block : BlocksTFC.getAllChestBlocks())
            ModelLoader.setCustomStateMapper(block, new StateMap.Builder().func_178442_a(BlockChest.field_176459_a).func_178441_a());
        ClientRegistry.bindTileEntitySpecialRenderer(TEChestTFC.class, new TESRChestTFC());

        ModelLoader.setCustomStateMapper(BlocksTFC.PIT_KILN, blockIn -> ImmutableMap.of(BlocksTFC.PIT_KILN.func_176223_P(), new ModelResourceLocation("tfc:empty")));
        ClientRegistry.bindTileEntitySpecialRenderer(TEPitKiln.class, new TESRPitKiln());

        ModelLoader.setCustomStateMapper(BlocksTFC.WORLD_ITEM, blockIn -> ImmutableMap.of(BlocksTFC.WORLD_ITEM.func_176223_P(), new ModelResourceLocation("tfc:empty")));
        ClientRegistry.bindTileEntitySpecialRenderer(TEWorldItem.class, new TESRWorldItem());

        ModelLoader.setCustomStateMapper(BlocksTFC.INGOT_PILE, blockIn -> ImmutableMap.of(BlocksTFC.INGOT_PILE.func_176223_P(), new ModelResourceLocation("tfc:empty")));
        ClientRegistry.bindTileEntitySpecialRenderer(TEIngotPile.class, new TESRIngotPile());

        for (Block block : BlocksTFC.getAllNormalItemBlocks().keySet())
            ModelLoader.setCustomModelResourceLocation(Item.func_150898_a(block), 0, new ModelResourceLocation(block.getRegistryName(), "normal"));

        for (Block block : BlocksTFC.getAllInventoryItemBlocks().keySet())
            ModelLoader.setCustomModelResourceLocation(Item.func_150898_a(block), 0, new ModelResourceLocation(block.getRegistryName(), "inventory"));

    }

    @SubscribeEvent
    @SideOnly(Side.CLIENT)
    public static void registerColorHandlerBlocks(ColorHandlerEvent.Block event)
    {
        BlockColors blockcolors = event.getBlockColors();

        blockcolors.func_186722_a((state, worldIn, pos, tintIndex) ->
                worldIn != null && pos != null ? BiomeColorHelper.func_180286_a(worldIn, pos) : ColorizerGrass.func_77480_a(0.5D, 1.0D),
            BlocksTFC.getAllBlockRockVariants().stream().filter(x -> x.type.isGrass).toArray(BlockRockVariant[]::new));

        blockcolors.func_186722_a((state, worldIn, pos, tintIndex) -> BlockFarmlandTFC.TINT[state.func_177229_b(BlockFarmlandTFC.MOISTURE)],
            BlocksTFC.getAllBlockRockVariants().stream().filter(x -> x.type == Rock.Type.FARMLAND).toArray(BlockRockVariant[]::new));

        blockcolors.func_186722_a((state, worldIn, pos, tintIndex) ->
                worldIn != null && pos != null ? BiomeColorHelper.func_180286_a(worldIn, pos) : ColorizerGrass.func_77480_a(0.5D, 1.0D),
            BlocksTFC.PEAT_GRASS);

        blockcolors.func_186722_a((state, worldIn, pos, tintIndex) ->
                worldIn != null && pos != null ? BiomeColorHelper.func_180288_c(worldIn, pos) : 0,
            BlocksTFC.getAllFluidBlocks().stream().filter(x -> x.func_176223_P().func_185904_a() == Material.field_151586_h).toArray(BlockFluidBase[]::new));

        blockcolors.func_186722_a((state, worldIn, pos, tintIndex) ->
                worldIn != null && pos != null ? BiomeColorHelper.func_180287_b(worldIn, pos) : ColorizerFoliage.func_77468_c(),
            BlocksTFC.getAllLeafBlocks().toArray(new Block[0]));
    }

    @SubscribeEvent
    @SideOnly(Side.CLIENT)
    @SuppressWarnings("deprecation")
    public static void registerColorHandlerItems(ColorHandlerEvent.Item event)
    {
        ItemColors itemColors = event.getItemColors();

        itemColors.func_186731_a((stack, tintIndex) ->
                event.getBlockColors().func_186724_a(((ItemBlock) stack.func_77973_b()).func_179223_d().func_176203_a(stack.func_77960_j()), null, null, tintIndex),
            BlocksTFC.getAllBlockRockVariants().stream().filter(x -> x.type.isGrass).toArray(BlockRockVariant[]::new));

        itemColors.func_186731_a((stack, tintIndex) ->
                event.getBlockColors().func_186724_a(((ItemBlock) stack.func_77973_b()).func_179223_d().func_176203_a(stack.func_77960_j()), null, null, tintIndex),
            BlocksTFC.PEAT_GRASS);

        itemColors.func_186731_a((stack, tintIndex) ->
                event.getBlockColors().func_186724_a(((ItemBlock) stack.func_77973_b()).func_179223_d().func_176203_a(stack.func_77960_j()), null, null, tintIndex),
            BlocksTFC.getAllLeafBlocks().stream().toArray(BlockLeavesTFC[]::new));

        itemColors.func_186730_a((stack, tintIndex) -> tintIndex == 1 ? EnumDyeColor.func_176766_a(stack.func_77952_i()).func_193350_e() : 0xFFFFFF,
            ItemsTFC.CERAMICS_UNFIRED_VESSEL_GLAZED, ItemsTFC.CERAMICS_FIRED_VESSEL_GLAZED);
    }

    /**
     * Turns "gem/diamond" + enum NORMAL into "gem/normal/diamond"
     */
    @SideOnly(Side.CLIENT)
    private static void registerEnumBasedMetaItems(String prefix, Enum e, Item item)
    {
        //noinspection ConstantConditions
        String registryName = item.getRegistryName().func_110623_a();
        StringBuilder path = new StringBuilder(MOD_ID).append(':');
        if (!Strings.isNullOrEmpty(prefix)) path.append(prefix).append('/');
        path.append(e.name());
        if (!Strings.isNullOrEmpty(prefix))
            path.append(registryName.replace(prefix, "")); // There well be a '/' at the start of registryName due to the prefix, so don't add an extra one.
        else path.append('/').append(registryName);
        ModelLoader.setCustomModelResourceLocation(item, e.ordinal(), new ModelResourceLocation(path.toString().toLowerCase()));
    }
}
